// 
// Module: ConcreteArfDataServer
// Author: Hassan Siddiqui
//
// Description: Concrete implementation of ArfDataServer class
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: ConcreteArfDataServer.h,v 1.7 2007/08/16 09:51:02 rsaxton Exp $

#ifndef ConcreteArfDataServer_H
#define ConcreteArfDataServer_H

/* General classes */
#include <iostream>

/* SAS-specific classes */
#include "CalServer.h"
#include "Dal.h"
#include "Tracer.h"
#include "DataSubspace.h"

/* Local Classes */
#include "DatafileAttributes.h"
#include "CoreData.h"
#include "GeneralIterator.h"
#include "PsfModel.h"
#include "PhotonDistribution.h"
#include "Frame.h"
#include "DetectorPosition.h"
#include "PixelIterator.h"
#include "EeCorrector.h"
#include "BadPixelCorrector.h"
#include "Ebins.h"
#include "SourcePosition.h"
#include "PixCoords.h"
#include "CamCoords.h"
#include "TelCoords.h"
#include "PosWCS.h"
#include "DetmapFilter.h"
#include "RegionExtent.h"
#include "RegionFunctions.h"
#include "CrossRegionArf.h"

/****************************************************
 *
 * class ConcreteArfDataServer
 *
 ***************************************************/

class ConcreteArfDataServer 
{
public:
  
  ConcreteArfDataServer(); 
  
  ~ConcreteArfDataServer();
  
  /********************************************
   *
   * General observation/mission/instrument info
   *
   ********************************************/

  Xmm::Instrument instrumentID()   const;


  /**************************
   *
   * Pointing info
   *
   **************************/
  
  /* Source position */
  double                     raref()                   const ;
  double                     decref()                  const ;

  bool                       useODF()                  const ;
  unsigned long              eventFlags()   	       const ;
  bool                       ignoreEdgePixels()        const ;
  bool                       ignoreOutofFOV()          const ;
  const unsigned long        ccdMask()                 const ;

  void                       readDSS()                       ;

  /********************************************
   *
   * Selection info
   *
   ********************************************/

  const DataSubspace * const       dataSubspace()          const;
  const RegionExtent * extent() const;

  RegionExtent * addDetRegion( PositionUtils * putils );


  /********************************************
   *
   * Ebins accessor
   *
   ********************************************/

  const BinnedAxis * const            ebins()   const; 

  /********************************************
   * 
   * Source Position
   *
   ********************************************/

  const SourcePosition * const  sourcePosition() const;

  /********************************************
   *
   * Misc
   *
   ********************************************/

  DatafileAttributes * specAttributes()  const;

  bool filterDSS ()    const;

  const string filteredTableName ()    const;

  bool keepFilteredTable ()    const ;

  bool isExtendedSource() const; 

  const PhotonDistribution * const  photonDistibution() const;

  const Frame * const frame()  const;
  
  PsfModel * const  psf() const;

  PixelIterator * const detmapPixelIterator();

  PixelIterator * const PsfPixelIterator();

  double framePeriod() const;

  bool setCcd( const CamCoords & camcoord);

  int ConcreteArfDataServer::backScale() const;
  
  int ConcreteArfDataServer::sourceBoxArea() const;
  
  bool ConcreteArfDataServer::crossRegion() const;

  void ConcreteArfDataServer::set_bpcorrectors();

  /*********************** 
   *  Conversion routines 
   ***********************/

  const DetectorPosition detxyToPixcoord( int detx, 
					  int dety) const;

  /* Convert from CAMCOORD2 to PIXCOORD */
  const PixCoords camCoord2ToPixCoord( const CamCoords & camcoords ) const;

  /* Convert from DETX/DETY (0.05 arcsec units) to camcoord2 */
  const CamCoords detxy005ToCamCoord2(int detx_005, 
				      int dety_005) const;

  /* Convert from CAMCOORD2 to TELCOORD */
  const TelCoords camCoord2ToTelCoord( const CamCoords & camCoords ) const;
 
  /* Return the positionUtils object */
  const PositionUtils * positionUtils() const;
  
  /* Convert from X/Y to DETX/DETY */
  SourcePosition xyToDetxy (double x, double y) const;

  /*********************** 
   *  Arf quantity routines 
   ***********************/

  /* These methods can be turned on/off by user */
  EffectiveAreaCurve effectiveAreaCurve  ( double theta, double phi ) const;

  CcdQuantumEfficiencyCurve quantumEfficiencyCurve ( double  x , 
						     double  y ) const;

  FilterTransmissionCurve filterTransmissionCurve ( double  theta , 
						    double  phi ) const;
  
 
  double pileupCorrection   (double  energy,
			     double  pixelFlux) const;

  const EeCorrector * const  eeCorrector () const;

  const BpCorrector * const  badPixCorrector () const;

  double ootCorrection ( ) const;
  
  GeoRegion PolyRegion () const;

private:

  void createDetmap();
  void checkRegionExtent();
  void geoExtent(GeoRegion &region);
 
  /* Member fields */

  DataSet              * _rmfset;
  Table                * _rmftable;

  DatafileAttributes   * _specAttributes;

  /* Detmap pointers */
  PixelIterator        * _rawDetmap;
  bool                 _filteredDetmapExists;
  PixelIterator        * _filteredDetmap;

  CoreData             * _coreData;

  Xmm::Instrument       _instrumentID;
  string               _instrument;
  string               _filter;
  string               _ppsvers;

  double               _raref;
  double               _decref;

  const PhotonDistribution   * _photonDistribution;
  const PsfModel             * _psf;
  const Frame                * _frame;

  bool                 _extendedsource;
  bool                 _modelEffectiveArea;
  bool                 _modelQuantumEfficiency;
  bool                 _modelFilterTransmission;
  bool                 _modelGratingAttenuation;
  bool                 _modelootcorr;
  bool                 _useodf;
  bool                 _crossRegion;
  
  const EeCorrector         * _eeCorrector; 
  const BpCorrector   * _badPixCorrector; 
  const BinnedAxis    * _ebins;

  DataSubspace  * _dss;

  const RegionExtent  * _extent;

  bool               _filterdss;
  string             _filteredtablename;
  bool               _keepfilteredtable;

  const EffectiveAreaCurve * _nulleacurve;
  const CcdQuantumEfficiencyCurve * _nullqecurve;
  const FilterTransmissionCurve   * _nullftcurve;

  const PosWCS *      _poswcs;

  const CoordPair *   _refcoords;

  PsfModel *          _bp_psf;

  DetmapFilter *      _detmapfilter;

  RegionFunctions *   _rfunction;
  
  CrossRegionArf * _crossRegionArf;

  unsigned            _sourceBoxArea;

  bool		      _regPosCoords;
  bool		      _regRawCoords;
  bool		      _TimOrBurstMode;
  bool		      _maskFound;
  

  Xmm::Mode           _windowMode;

  unsigned long       _flag;
  unsigned long       _ccdMask;

};

#endif
