//
// Module: CrossRegionArf
// Author: Richard Saxton
//
// Description: 
//
// Sets up a PSF model for a circular source box.
// It is valid for a circular source box centred on the source.
//
// Modifications: See ChangeLog
// 
// $Id: CrossRegionArf.cc,v 1.3 2007/06/15 10:20:36 rsaxton Exp $

// General classes
#include <iostream>
#include <cmath>

// SAS-specific classes

// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "NumericalRange.h"
#include "EncircledEnergyDataServer.h"
#include "DetmapFilter.h"

// Local classes
#include "CrossRegionArf.h"

/* Constructor */
CrossRegionArf::CrossRegionArf(string cr_spectrumsetname,
                                    PixelIterator *rawDetmap,
                                    const PositionUtils * const putils) :
 _rawDetmap(rawDetmap),
 _putils(putils),
 _init(false)                                                                   
{
  PROC;

   errstr 
    << "Opening spectrumset " << cr_spectrumsetname << "..."
    << Sas::message(Msg::App, Msg::Sparse);
  
   DataSet * cr_spectrumset = dataSetServer->open( cr_spectrumsetname,
                         DataSet::Read );
  
  /* Access spectrum  */
   Table * spectrumtable = cr_spectrumset->table("SPECTRUM"); 
  
  /* Read the spectrum attributes */
   _crdss = new DataSubspace(spectrumtable);
  
  /* Read the spectrum attributes */
   _crspecAttributes = new DatafileAttributes(cr_spectrumsetname); 
  
   errstr 
    << "CrossRegionArf: constructed"
    << Sas::message(Msg::App, Msg::Verbose);


  ENDPROC;
};

CrossRegionArf::~CrossRegionArf()
{
  PROC;
  
  // Here delete all the pointers //
  
  errstr 
    << "CrossRegionArf destroyed."
    << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;

}

/* Method: init() - sets up filtered detmap and a PSF for each pixel */
void CrossRegionArf::init()
{
  PROC;

  /* If not yet set up, create a detmap filtered by the second region 
     and a set of PSFs and effective area curves */
    if (! _init ) {

      /* Use initial psf photon energy of 2.0 keV, but this will be
         reset later */
        double psfenergy = 2000.0;
    
      /* Get the effective area server */
       EffectiveAreaDataServer * effds = calServer.getAtom(effds);
 
      /* Filter the input detector map */          	
       _crfilteredDetmap = detmapPixelIterator();
       
      /* Loop over each pixel in the filtered detector map */
       FOREACH(_crfilteredDetmap)
        {
          PixelElement pixel = _crfilteredDetmap->normalisedItem();
         
         /* Create a PSF at this position */
          _crPsf.push_back(new CalPsfModel(psfenergy,
                pixel.xCentre(), 
                pixel.yCentre(),
                EXTENDED,
                _putils));
                
 
         /* Find the telescope coords of this pixel */ 
          CamCoords offset = _putils->detxy005ToCamCoord2( 
                                       pixel.xCentre(), pixel.yCentre());
                                                                                
          TelCoords t = _putils->camCoord2ToTelCoord(offset);
                        
         /* Find the effective area at this telescope position */
          _eacurve.push_back(effds->effectiveAreaCurve(t.theta(), t.phi()));  

         /* Set an array of normalisations */
          _pixnorm.push_back(pixel.value());
 
        /* Set the pixel width - note implicit assumption here that the
           source and second region pixels are same size */
          _x_Width=pixel.xWidth(); 
          _y_Width=pixel.yWidth();
          
        }   
 
      /* Get the energies from the curve */
        _energyVector =  _eacurve[0].abscissa();  
                 
      /* Set flag so this code is not repeated */
        _init=true; 
    }
 
  ENDPROC;
    
} // init
       
/* Method: detmapPixelIterator() - sets up the filtered detector map */
PixelIterator * const  CrossRegionArf::detmapPixelIterator()
{
  
  PROC;

  /* Create a filtered detmap using the region from
     the second spectral file */
    DetmapFilter * detmapfilter = new DetmapFilter;
       
    //filteredDetmap = detmapfilter->filteredMap(_rawDetmap, 
      //        "cr_filteredpixellist.ds:EVENTS", _crspecAttributes, 
        //      _crdss, false);
 
    RETURN detmapfilter->filteredMap(_rawDetmap, 
              "cr_filteredpixellist.ds:EVENTS", _crspecAttributes, 
              _crdss, false);
              
}

/* Method: vigPsf(double theta, double phi)
 * Purpose: Calculate an energy-dependent curve of effective area (+ vignetting)
            times the fraction of the PSF which intersects the source region 
*/
EffectiveAreaCurve CrossRegionArf::psfVignet( double theta_out,
                                              double phi_out )
{ 
  PROC;

    /* Initialise the detector map and PSF functions */
     this->init();

    //EffectiveAreaCurve eaout;
     vector<double> eaout(_energyVector.size());
     
    /* Convert TelCoords back to detector coords */
     //TelCoords * t = new TelCoords(theta_out, phi_out);
     TelCoords t(theta_out, phi_out);
     CamCoords camcoords = _putils->TelCoordTocamCoord2(t);
     double x_out=_putils->CamCoord2Todetxy005(camcoords.x());
     double y_out=_putils->CamCoord2Todetxy005(camcoords.y());
  
    /* Loop over each energy */
     for (unsigned int e = 0; e <_energyVector.size(); ++e) { 
     
       double total=0.0;
     
      /* Loop over each pixel in the filtered detector map */
       for (unsigned int i=0; i<_crfilteredDetmap->size(); i++)
       {
          /* Set the PSF energy */
            _crPsf[i]->energy(_energyVector[e]);
            
          /* Get the PSF fraction at this energy which falls within the
             output bin - NB: implicit assumption here that input and output
             detector maps have the same size pixels */
            double psfbit = _crPsf[i]->int_psf(x_out, y_out, _x_Width, _y_Width);
            
          /* Add the product of the eff area at the input pixel position and the
             PSF contribution in the output pixel, weighted by the normalised
             counts value at this input position */ 
            total += psfbit * _eacurve[i].area(_energyVector[e]) * _pixnorm[i];
            
            //if (e==100 && i==1)
              // cout<<"Bits: "<<psfbit<<" "<<_eacurve[i].area(_energyVector[e])
                //   <<" "<<_pixnorm[i]<<endl;
                       
      }      

     /* Update the output area array for this energy */
      eaout[e]=total;
     
    } // Energy loop

    RETURN Slatec::CubicHermiteInterpolant<double>(_energyVector, eaout);
  
}

