// 
// Module: OffCCDPixelIterator
// Author: Richard Saxton 
//
// Description: 
//
// An iterator for pixels which lie outside the live CCD area
//
// Modifications: See ChangeLog
// 
// $Id: OffCCDPixelIterator.cc,v 1.8 2007/11/05 09:04:05 rsaxton Exp $

// General classes
#include <iostream>
#include <cmath>

// SAS-specific classes
#include "Dal.h"
#include "Region.h"
#include "ShapeScanner.h"
#include "CamCoords.h"
#include "GeometryDataServer.h"
#include "CalServer.h"
#include "Xmm.h"
#include "EventAttributes.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"

// Local classes
#include "ArfDataServer.h"
#include "OffCCDPixelIterator.h"
#include "RegionExtent.h"
#include "PositionUtils.h"

/* Constructors */
OffCCDPixelIterator::OffCCDPixelIterator(const GeoRegion & sourceArea,
                                         const RegionExtent * extent,
                                         const PositionUtils * const putils,
                                         unsigned long flags,
                                         bool ignoreoutofov,
                                         unsigned long ccdmask) :
  _firstcalled(false),
  _offccdelements(0),
  _totalelements(0)
  
{
  PROC;
  
  using namespace scanningLib;

 /* Set local variables for the pixel selection options */
//  bool ignoreEdgePixels = arfDataServer->ignoreEdgePixels();
//  unsigned long ccdmask = arfDataServer->ccdMask();

 /* Define the width of each grid element in detector coords 
    Use the nominal grid size of 1 arcsec for large areas (rad>2 amin)
    and 0.5 arcsecs for small (rad<2 amin) */
  float halfwidth = (extent->ymax()-extent->ymin()) / 2.0;
  if (halfwidth < 2400.0) {
     _gridWidth=10.0;
  } else if (halfwidth < 4800.0) {
     _gridWidth=20.0;
  } else if (halfwidth < 9600.0) {
     _gridWidth=40.0;
  } else if (halfwidth < 16800.0) {
     _gridWidth=80.0;
  } else {
      _gridWidth=160.0;
  }

 /* Set up grid structure to evaluate which pixels lie within the
      source region. Work in detector coords. */
  struct Grid yGrid(_gridWidth, extent->ymin()+_gridWidth/2.0), 
              xGrid(_gridWidth, extent->xmin()+_gridWidth/2.0);
  //yGrid._origin = extent->ymin();
  //yGrid._stride = _gridWidth;
  //yGrid._length = UNT((extent->ymax()-extent->ymin()+1.0)/ _gridWidth);
  //xGrid._origin = extent->xmin();
  //xGrid._stride = _gridWidth;
  //xGrid._length = UNT((extent->xmax()-extent->xmin()+1.0)/ _gridWidth);

  errstr<<"Grid origin: "<<extent->ymin()
    <<" "<<UNT((extent->ymax()-extent->ymin()+1.0)/ _gridWidth)
    <<" "<<extent->xmin()
    <<" "<<UNT((extent->xmax()-extent->xmin()+1.0)/ _gridWidth)
    <<" stride: "<<_gridWidth
    <<" xmin: "<<xGrid.origin<<" ymin: "<<yGrid.origin
             << Sas::message(Msg::App, Msg::Verbose);
        // <<" length: "<<xGrid._length<<" "<<yGrid._length

 /* Define structures to hold region information */
  list<Span> spans;
  list<Span>::const_iterator iter;
  struct RegionScanner scanner( sourceArea, xGrid, yGrid );

 /* Loop over each grid element within these spans */
  int col=0;
  double detx=0.0, dety=0.0;
  int iy=0;
  CamCoords testpixel(0.0,0.0,0.0);
  PixCoords pixcoord(0,0);

  GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

 /* Get window object for this observing mode */
  ModeParameters *modeParam = 0; modeParam = calServer.getAtom(modeParam);
  CCDwindow *ccdwindow = new CCDwindow(modeParam->windowX0(),
                                         modeParam->windowY0(),
                        (modeParam->windowX0()+modeParam->windowDX()-1),
                        (modeParam->windowY0()+modeParam->windowDY()-1));

  errstr<<"Set up window of size: "<<modeParam->windowX0()<<" "
        << modeParam->windowY0()<<" "
        <<(modeParam->windowX0()+modeParam->windowDX()-1)<<" "
        <<(modeParam->windowY0()+modeParam->windowDY()-1)
        << Sas::message(Msg::App, Msg::Noisy);

 /* Get CCD size */
  unsigned minx = calServer.instrument()->minX();
  unsigned maxx = calServer.instrument()->maxX();
  unsigned miny = calServer.instrument()->minY();
  unsigned maxy = calServer.instrument()->maxY();

  errstr<<"CCD size - Min/max is :"<<minx<<" "<<maxx<<" "<<miny<<" "<<maxy
             << Sas::message(Msg::App, Msg::Noisy);

/* Get the number of the central CCD */
  unsigned cent_ccd = calServer.instrument()->focusCcd();

 /* Loop over each column (X) */
  for (scanner.reset(); scanner.next(col, spans);) {

    detx = xGrid[col];

    for (iter = spans.begin(); iter !=spans.end(); ++iter) {

   /* Loop over each Y grid element within this span */
     for (iy=iter->min; iy<=iter->max; iy++) {

       _totalelements++;

       dety = yGrid[iy];
       testpixel = putils->detxy005ToCamCoord2(detx, dety);
       pixcoord  = putils->camCoord2ToPixCoord(testpixel);

   /* Find which CCD this is on and set in CAL, then need to reget the
      GeometryDataServer - otherwise out of fOV check doesn't work */
       Ccd ccd = ct->onCcd(testpixel.x(), testpixel.y());
       if (ccd > 0) {
           calServer.state()->ccd()->set(ccd);
       }
                                                                                
       GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

   /* Check for outside FOV */
       CalInt16Vector chipX(0, 1), chipY(0, 1);
       chipX[0] = pixcoord.x();
       chipY[0] = pixcoord.y();
       XmmPixelList chipCoord1 = ct->pixCoord1List(chipX, chipY).toChipCoord();

       bool infov = chipCoord1.insideFOV()[0];

   /* Check if pixel is outside an observing window. Normally this only
      applies to the central CCD but all chips are affected in PN large 
      window mode */
       bool insideWindow=true;

       if ( (ccd == cent_ccd || ( calServer.instrument()->id() == Xmm::EPN &&
            calServer.state()->mode()->get() == Xmm::PRIME_LARGE_WINDOW ) )
            && ! ( ccdwindow->isPixelInside(
                        PixelLocation(pixcoord.x(),pixcoord.y()))) )
             {insideWindow=false;}

       errstr<<"Test pixel is: "<<iy<<" "<<col<<" "<<" det: "<<detx<<" "<<dety
             <<"Raw: "<<pixcoord.x()<<" "<<pixcoord.y()
             <<" Inside FOV: "<<infov<<" "<<ignoreoutofov
             <<" ccd = "<<ccd<<" central ccd = "<<cent_ccd
             <<" inside window: "<<insideWindow
             << Sas::message(Msg::App, Msg::Noisy);

   /* Add pixels which dont fall on selected CCDs into off_ccd vector.
      If requested, also add edge-of-chip pixels into vector */

       if (  (! (ccdmask & 1<<ccd) ) || (! insideWindow) || 
                                        (! infov && ignoreoutofov) ||
          // Next to CCD border
         ( ! (flags & 1<<EventAttributes::CLOSE_TO_CCD_BORDER) && 
            (pixcoord.x() == minx || pixcoord.x() == maxx
                || pixcoord.y() == miny || pixcoord.y() == maxy) ) )
       {
        _OffCCDPixel.push_back(PixelElement(col,
                                            iy,
                                            detx, dety,
                                            _gridWidth, _gridWidth,
                                            1.0));
        _offccdelements++;

          errstr<<"OFF CCD: "<<testpixel.x()<<" "<<testpixel.y()
                <<" Det: "<<detx<<" "<<dety
                <<" Pixel: "<<pixcoord.x()<<" "<<pixcoord.y()
             << Sas::message(Msg::App, Msg::Noisy);

       }
       errstr<<"CCD: "<<ccd
           <<"Det: "<<detx<<" "<<dety
                <<" Pixel: "<<pixcoord.x()<<" "<<pixcoord.y()
           << Sas::message(Msg::App, Msg::Noisy);
    }
   }
  }

  /* Set up the WCS convertor to pass widths */
  //_wcsconv = new WCSConvertor("DETX", 0, 0, _gridWidth,
    //                             "DETY", 0, 0,_gridWidth);
  _wcsconv = new WCSConvertor("DETX", col-1, detx, _gridWidth,
                               "DETY", iy-1, dety, _gridWidth);

  ENDPROC;
}

/* Destructor */
OffCCDPixelIterator::~OffCCDPixelIterator()
{
  PROC;
  ENDPROC;
}

/* Iterator methods */
void OffCCDPixelIterator::first()
{
  _firstcalled = true;
  _index=0;
}


void OffCCDPixelIterator::next ()
{
  if (isDone())
    {
      errstr<<"Anomaly in OffCCDPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  _index++;
}

bool OffCCDPixelIterator::isDone()
{
  PROC;
    if (!_firstcalled)
    {
      errstr<<"Anomaly in OffCCDPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  RETURN (_index >= _offccdelements);
}


PixelElement OffCCDPixelIterator::currentItem()
{
  PROC;

    if (!_firstcalled)
    {
      errstr<<"Anomaly in OffCCDPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  RETURN _OffCCDPixel[_index];
}

unsigned OffCCDPixelIterator::size() const
{
  PROC;
  RETURN _offccdelements;
}

unsigned OffCCDPixelIterator::xbins() const
{
  PROC;
  
  RETURN _offccdelements;
}

unsigned OffCCDPixelIterator::totalPixels() const
{
  PROC;
  
  RETURN UNT(_totalelements * _gridWidth * _gridWidth);
}

/* WCSConvertor object */
const WCSConvertor * OffCCDPixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}


